'''Test the serial worker'''

# Internal imports
from common import TestQless

import time
from threading import Thread

# The stuff we're actually testing
from qless import logger
from qless.workers.serial import SerialWorker


class SerialJob(object):
    '''Dummy class'''
    @staticmethod
    def foo(job):
        '''Dummy job'''
        time.sleep(job.data.get('sleep', 0))
        try:
            job.complete()
        except:
            logger.exception('Unable to complete job %s' % job.jid)


class Worker(SerialWorker):
    '''A worker that limits the number of jobs it runs'''
    def jobs(self):
        '''Yield only a few jobs'''
        generator = SerialWorker.jobs(self)
        for _ in xrange(5):
            yield generator.next()

    def kill(self, jid):
        '''We'll push a message to redis instead of falling on our sword'''
        self.client.redis.rpush('foo', jid)
        raise KeyboardInterrupt()

    def signals(self):
        '''Do not set any signal handlers'''
        pass


class NoListenWorker(Worker):
    '''A worker that just won't listen'''
    def listen(self, _):
        '''Don't listen for lost locks'''
        pass


class TestWorker(TestQless):
    '''Test the worker'''
    def setUp(self):
        TestQless.setUp(self)
        self.queue = self.client.queues['foo']
        self.thread = None

    def tearDown(self):
        if self.thread:
            self.thread.join()
        TestQless.tearDown(self)

    def test_basic(self):
        '''Can complete jobs in a basic way'''
        jids = [self.queue.put(SerialJob, {}) for _ in xrange(5)]
        NoListenWorker(['foo'], self.client, interval=0.2).run()
        states = [self.client.jobs[jid].state for jid in jids]
        self.assertEqual(states, ['complete'] * 5)

    def test_jobs(self):
        '''The jobs method yields None if there are no jobs'''
        worker = NoListenWorker(['foo'], self.client, interval=0.2)
        self.assertEqual(worker.jobs().next(), None)

    def test_sleeps(self):
        '''Make sure the client sleeps if there aren't jobs to be had'''
        for _ in xrange(4):
            self.queue.put(SerialJob, {})
        before = time.time()
        NoListenWorker(['foo'], self.client, interval=0.2).run()
        self.assertGreater(time.time() - before, 0.2)

    def test_lost_locks(self):
        '''The worker should be able to stop processing if need be'''
        jid = [self.queue.put(SerialJob, {'sleep': 0.1}) for _ in xrange(5)][0]
        self.thread = Thread(
            target=Worker(['foo'], self.client, interval=0.2).run)
        self.thread.start()
        # Now, we'll timeout one of the jobs and ensure that kill is invoked
        while self.client.jobs[jid].state != 'running':
            time.sleep(0.01)
        self.client.jobs[jid].timeout()
        self.assertEqual(self.client.redis.brpop('foo', 1), ('foo', jid))

    def test_kill(self):
        '''Should be able to fall on its sword if need be'''
        worker = SerialWorker([], self.client)
        worker.jid = 'foo'
        thread = Thread(target=worker.kill, args=(worker.jid,))
        thread.start()
        thread.join()
        self.assertFalse(thread.is_alive())

    def test_kill_dead(self):
        '''If we've moved on to another job, say so'''
        # If this tests runs to completion, it has succeeded
        worker = SerialWorker([], self.client)
        worker.kill('foo')

    def test_shutdown(self):
        '''We should be able to shutdown a serial worker'''
        # If this test finishes, it passes
        worker = SerialWorker([], self.client, interval=0.1)
        worker.stop()
        worker.run()



#! /usr/bin/env python

'''Some exception classes'''


class QlessException(Exception):
    '''Any and all qless exceptions'''
    pass


class LostLockException(QlessException):
    '''Lost lock on a job'''
    pass



'''Test worker'''

# Internal imports
from common import TestQless

import qless
from qless.workers import Worker

# External dependencies
import os
import itertools


class TestWorker(TestQless):
    '''Test the worker'''
    def setUp(self):
        TestQless.setUp(self)
        self.worker = Worker(['foo'], self.client)

    def test_proctitle(self):
        '''Make sure we can get / set the process title'''
        before = Worker.title()
        Worker.title('Foo')
        self.assertNotEqual(before, Worker.title())

    def test_kill(self):
        '''The base worker class' kill method should raise an exception'''
        self.assertRaises(NotImplementedError, self.worker.kill, 1)

    def test_clean(self):
        '''Should be able to clean a directory'''
        if not os.path.exists('test/tmp'):
            os.makedirs('test/tmp')
        self.assertEqual(os.listdir('test/tmp'), [])
        os.makedirs('test/tmp/foo/bar')
        with open('test/tmp/file.out', 'w+'):
            pass
        self.assertNotEqual(os.listdir('test/tmp'), [])
        Worker.clean('test/tmp')
        self.assertEqual(os.listdir('test/tmp'), [])

    def test_sandbox(self):
        '''The sandbox utility should work'''
        path = 'test/tmp/foo'
        self.assertFalse(os.path.exists(path))
        try:
            with Worker.sandbox(path):
                self.assertTrue(os.path.exists(path))
                for name in ['whiz', 'widget', 'bang']:
                    with open(os.path.join(path, name), 'w+'):
                        pass
                # Now raise an exception
                raise ValueError('foo')
        except ValueError:
            pass
        # Make sure the directory has been cleaned
        self.assertEqual(os.listdir(path), [])
        os.rmdir(path)

    def test_sandbox_exists(self):
        '''Sandbox creation should not throw an error if the path exists'''
        path = 'test/tmp'
        self.assertEqual(os.listdir(path), [])
        with Worker.sandbox(path):
            pass
        # If we get to this point, the test succeeds
        self.assertTrue(True)

    def test_dirty_sandbox(self):
        '''If a sandbox is dirty on arrival, clean it first'''
        path = 'test/tmp/foo'
        with Worker.sandbox(path):
            for name in ['whiz', 'widget', 'bang']:
                with open(os.path.join(path, name), 'w+'):
                    pass
            # Now it's sullied. Clean it up
            self.assertNotEqual(os.listdir(path), [])
            with Worker.sandbox(path):
                self.assertEqual(os.listdir(path), [])
        os.rmdir(path)

    def test_resume(self):
        '''We should be able to resume jobs'''
        queue = self.worker.client.queues['foo']
        queue.put('foo', {})
        job = self.worker.jobs().next()
        self.assertTrue(isinstance(job, qless.Job))
        # Now, we'll create a new worker and make sure it gets that job first
        worker = Worker(['foo'], self.client, resume=[job])
        self.assertEqual(worker.jobs().next().jid, job.jid)

    def test_unresumable(self):
        '''If we can't heartbeat jobs, we should not try to resume it'''
        queue = self.worker.client.queues['foo']
        queue.put('foo', {})
        # Pop from another worker
        other = qless.Client(hostname='other')
        job = other.queues['foo'].pop()
        self.assertTrue(isinstance(job, qless.Job))
        # Now, we'll create a new worker and make sure it gets that job first
        worker = Worker(
            ['foo'], self.client, resume=[self.client.jobs[job.jid]])
        self.assertEqual(worker.jobs().next(), None)

    def test_resumable(self):
        '''We should be able to find all the jobs that can be resumed'''
        # We're going to put some jobs into some queues, and pop them.
        jid = self.client.queues['foo'].put('Foo', {})
        self.client.queues['bar'].put('Foo', {})
        self.client.queues['foo'].pop()
        self.client.queues['bar'].pop()

        # Now, we should be able to see a resumable job in 'foo', but we should
        # not see the job that we popped from 'bar'
        worker = Worker(['foo'], self.client, resume=True)
        jids = [job.jid for job in worker.resume]
        self.assertEqual(jids, [jid])

    def test_divide(self):
        '''We should be able to divide resumable jobs evenly'''
        items = self.worker.divide(range(100), 7)
        # Make sure we have the same items as output as input
        self.assertEqual(sorted(itertools.chain(*items)), range(100))
        lengths = [len(batch) for batch in items]
        self.assertLessEqual(max(lengths) - min(lengths), 1)



'''A worker that serially pops and complete jobs'''

import os
import time

from . import Worker


class SerialWorker(Worker):
    '''A worker that just does serial work'''
    def __init__(self, *args, **kwargs):
        Worker.__init__(self, *args, **kwargs)
        # The jid that we're working on at the moment
        self.jid = None
        # This is the sandbox we use
        self.sandbox = kwargs.pop(
            'sandbox', os.path.join(os.getcwd(), 'qless-py-workers'))

    def kill(self, jid):
        '''The best way to do this is to fall on our sword'''
        if jid == self.jid:
            exit(1)

    def run(self):
        '''Run jobs, popping one after another'''
        # Register our signal handlers
        self.signals()

        with self.listener():
            for job in self.jobs():
                # If there was no job to be had, we should sleep a little bit
                if not job:
                    self.jid = None
                    self.title('Sleeping for %fs' % self.interval)
                    time.sleep(self.interval)
                else:
                    self.jid = job.jid
                    self.title('Working on %s (%s)' % (job.jid, job.klass_name))
                    with Worker.sandbox(self.sandbox):
                        job.sandbox = self.sandbox
                        job.process()
                if self.shutdown:
                    break



'''All our configuration operations'''

import simplejson as json


class Config(object):
    '''A class that allows us to change and manipulate qless config'''
    def __init__(self, client):
        self._client = client

    def __getattr__(self, attr):
        if attr == 'all':
            return json.loads(self._client('config.get'))
        raise AttributeError('qless.Config has no attribute %s' % attr)

    def __len__(self):
        return len(self.all)

    def __getitem__(self, option):
        result = self._client('config.get', option)
        if not result:
            return None
        return json.loads(result)

    def __setitem__(self, option, value):
        return self._client('config.set', option, value)

    def __delitem__(self, option):
        return self._client('config.unset', option)

    def __contains__(self, option):
        return dict.__contains__(self.all, option)

    def __iter__(self):
        return iter(self.all)

    def clear(self):
        '''Remove all keys'''
        for key in self.all.keys():
            self._client('config.unset', key)

    def get(self, option, default=None):
        '''Get a particular option, or the default if it's missing'''
        val = self[option]
        return ((val == None) and default) or val

    def items(self):
        '''Just like `dict.items`'''
        return self.all.items()

    def keys(self):
        '''Just like `dict.keys`'''
        return self.all.keys()

    def pop(self, option, default=None):
        '''Just like `dict.pop`'''
        val = self[option]
        del self[option]
        return ((val == None) and default) or val

    def update(self, other=(), **kwargs):
        '''Just like `dict.update`'''
        _kwargs = dict(kwargs)
        _kwargs.update(other)
        for key, value in _kwargs.items():
            self[key] = value

    def values(self):
        '''Just like `dict.values`'''
        return self.all.values()



'''Test the serial worker'''

# Internal imports
from common import TestQless

import time
import gevent

# The stuff we're actually testing
from qless.workers.greenlet import GeventWorker


class GeventJob(object):
    '''Dummy class'''
    @staticmethod
    def foo(job):
        '''Dummy job'''
        job.data['sandbox'] = job.sandbox
        job.complete()


class PatchedGeventWorker(GeventWorker):
    '''A worker that limits the number of jobs it runs'''
    @classmethod
    def patch(cls):
        '''Don't monkey-patch anything'''
        pass

    def jobs(self):
        '''Yield only a few jobs'''
        generator = GeventWorker.jobs(self)
        for _ in xrange(5):
            yield generator.next()

    def listen(self, _):
        '''Don't actually listen for pubsub events'''
        pass


class TestWorker(TestQless):
    '''Test the worker'''
    def setUp(self):
        TestQless.setUp(self)
        self.worker = PatchedGeventWorker(
            ['foo'], self.client, greenlets=1, interval=0.2)
        self.queue = self.client.queues['foo']
        self.thread = None

    def tearDown(self):
        if self.thread:
            self.thread.join()
        TestQless.tearDown(self)

    def test_basic(self):
        '''Can complete jobs in a basic way'''
        jids = [self.queue.put(GeventJob, {}) for _ in xrange(5)]
        self.worker.run()
        states = [self.client.jobs[jid].state for jid in jids]
        self.assertEqual(states, ['complete'] * 5)
        sandboxes = [self.client.jobs[jid].data['sandbox'] for jid in jids]
        for sandbox in sandboxes:
            self.assertIn('qless-py-workers/greenlet-0', sandbox)

    def test_sleeps(self):
        '''Make sure the client sleeps if there aren't jobs to be had'''
        for _ in xrange(4):
            self.queue.put(GeventJob, {})
        before = time.time()
        self.worker.run()
        self.assertGreater(time.time() - before, 0.2)

    def test_kill(self):
        '''Can kill greenlets when it loses its lock'''
        worker = PatchedGeventWorker(['foo'], self.client)
        greenlet = gevent.spawn(gevent.sleep, 1)
        worker.greenlets['foo'] = greenlet
        worker.kill('foo')
        greenlet.join()
        self.assertIsInstance(greenlet.value, gevent.GreenletExit)

    def test_kill_dead(self):
        '''Does not panic if the greenlet handling a job is no longer around'''
        # This test succeeds if it finishes without an exception
        self.worker.kill('foo')



#! /usr/bin/env python

'''Both the regular Job and RecurringJob classes'''

import os
import time
import types
import traceback
import simplejson as json

# Internal imports
from qless import logger
from qless.exceptions import LostLockException, QlessException


class BaseJob(object):
    '''This is a dictionary of all the classes that we've seen, and
    the last load time for each of them. We'll use this either for
    the debug mode or the general mechanism'''
    _loaded = {}

    def __init__(self, client, **kwargs):
        self.client = client
        for att in ['jid', 'priority']:
            object.__setattr__(self, att, kwargs[att])
        object.__setattr__(self, 'klass_name', kwargs['klass'])
        object.__setattr__(self, 'queue_name', kwargs['queue'])
        # Because of how Lua parses JSON, empty tags comes through as {}
        object.__setattr__(self, 'tags', kwargs['tags'] or [])
        object.__setattr__(self, 'data', json.loads(kwargs['data']))

    def __setattr__(self, key, value):
        if key == 'priority':
            return self.client('priority', self.jid, value
                ) and object.__setattr__(self, key, value)
        else:
            return object.__setattr__(self, key, value)

    def __getattr__(self, key):
        if key == 'queue':
            # An actual queue instance
            object.__setattr__(self, 'queue',
                self.client.queues[self.queue_name])
            return self.queue
        elif key == 'klass':
            # Get a reference to the provided klass
            object.__setattr__(self, 'klass', self._import(self.klass_name))
            return self.klass
        raise AttributeError('%s has no attribute %s' % (
            self.__class__.__module__ + '.' + self.__class__.__name__, key))

    @staticmethod
    def reload(klass):
        '''Force a reload of this klass on next import'''
        BaseJob._loaded[klass] = 0

    @staticmethod
    def _import(klass):
        '''1) Get a reference to the module
           2) Check the file that module's imported from
           3) If that file's been updated, force a reload of that module
                return it'''
        mod = __import__(klass.rpartition('.')[0])
        for segment in klass.split('.')[1:-1]:
            mod = getattr(mod, segment)

        # Alright, now check the file associated with it. Note that clases
        # defined in __main__ don't have a __file__ attribute
        if klass not in BaseJob._loaded:
            BaseJob._loaded[klass] = time.time()
        if hasattr(mod, '__file__'):
            mtime = os.stat(mod.__file__).st_mtime
            if BaseJob._loaded[klass] < mtime:
                mod = reload(mod)

        return getattr(mod, klass.rpartition('.')[2])

    def cancel(self):
        '''Cancel a job. It will be deleted from the system, the thinking
        being that if you don't want to do any work on it, it shouldn't be in
        the queueing system.'''
        return self.client('cancel', self.jid)

    def tag(self, *tags):
        '''Tag a job with additional tags'''
        return self.client('tag', 'add', self.jid, *tags)

    def untag(self, *tags):
        '''Remove tags from a job'''
        return self.client('tag', 'remove', self.jid, *tags)


class Job(BaseJob):
    '''The Job class'''
    def __init__(self, client, **kwargs):
        BaseJob.__init__(self, client, **kwargs)
        self.client = client
        for att in ['state', 'tracked', 'failure',
            'history', 'dependents', 'dependencies']:
            object.__setattr__(self, att, kwargs[att])

        # The reason we're using object.__setattr__ directly is because
        # we have __setattr__ defined for this class, and we're actually
        # just interested in setting these memebers directly
        object.__setattr__(self, 'expires_at', kwargs['expires'])
        object.__setattr__(self, 'original_retries', kwargs['retries'])
        object.__setattr__(self, 'retries_left', kwargs['remaining'])
        object.__setattr__(self, 'worker_name', kwargs['worker'])
        # Because of how Lua parses JSON, empty lists come through as {}
        object.__setattr__(self, 'dependents', kwargs['dependents'] or [])
        object.__setattr__(self, 'dependencies', kwargs['dependencies'] or [])

    def __getattr__(self, key):
        if key == 'ttl':
            # How long until this expires, in seconds
            return self.expires_at - time.time()
        return BaseJob.__getattr__(self, key)

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __repr__(self):
        return '<%s %s>' % (self.klass_name, self.jid)

    def process(self):
        '''Load the module containing your class, and run the appropriate
        method. For example, if this job was popped from the queue
        ``testing``, then this would invoke the ``testing`` staticmethod of
        your class.'''
        try:
            method = getattr(self.klass, self.queue_name,
                getattr(self.klass, 'process', None))
        except Exception as exc:
            # We failed to import the module containing this class
            logger.exception('Failed to import %s', self.klass_name)
            return self.fail(self.queue_name + '-' + exc.__class__.__name__,
                'Failed to import %s' % self.klass_name)

        if method:
            if isinstance(method, types.FunctionType):
                try:
                    logger.info('Processing %s in %s',
                        self.jid, self.queue_name)
                    method(self)
                    logger.info('Completed %s in %s',
                        self.jid, self.queue_name)
                except Exception as exc:
                    # Make error type based on exception type
                    logger.exception('Failed %s in %s: %s',
                        self.jid, self.queue_name, repr(method))
                    self.fail(self.queue_name + '-' + exc.__class__.__name__,
                        traceback.format_exc())
            else:
                # Or fail with a message to that effect
                logger.error('Failed %s in %s : %s is not static',
                    self.jid, self.queue_name, repr(method))
                self.fail(self.queue_name + '-method-type',
                    repr(method) + ' is not static')
        else:
            # Or fail with a message to that effect
            logger.error('Failed %s : %s is missing a method "%s" or "process"',
                         self.jid, self.klass_name, self.queue_name)
            self.fail(self.queue_name + '-method-missing', self.klass_name +
                ' is missing a method "' + self.queue_name + '" or "process"')

    def move(self, queue, delay=0, depends=None):
        '''Move this job out of its existing state and into another queue. If
        a worker has been given this job, then that worker's attempts to
        heartbeat that job will fail. Like ``Queue.put``, this accepts a
        delay, and dependencies'''
        logger.info('Moving %s to %s from %s',
            self.jid, queue, self.queue_name)
        return self.client('put', queue, self.jid, self.klass_name,
            json.dumps(self.data), delay, 'depends', json.dumps(depends or [])
        )

    def complete(self, nextq=None, delay=None, depends=None):
        '''Turn this job in as complete, optionally advancing it to another
        queue. Like ``Queue.put`` and ``move``, it accepts a delay, and
        dependencies'''
        if nextq:
            logger.info('Advancing %s to %s from %s',
                self.jid, nextq, self.queue_name)
            return self.client('complete', self.jid, self.client.worker_name,
                self.queue_name, json.dumps(self.data), 'next', nextq,
                'delay', delay or 0, 'depends', json.dumps(depends or [])
            ) or False
        else:
            logger.info('Completing %s', self.jid)
            return self.client('complete', self.jid, self.client.worker_name,
                self.queue_name, json.dumps(self.data)) or False

    def heartbeat(self):
        '''Renew the heartbeat, if possible, and optionally update the job's
        user data.'''
        logger.debug('Heartbeating %s (ttl = %s)', self.jid, self.ttl)
        try:
            self.expires_at = float(self.client('heartbeat', self.jid,
            self.client.worker_name, json.dumps(self.data)) or 0)
        except QlessException:
            print 'Raising exception'
            raise LostLockException(self.jid)
        logger.debug('Heartbeated %s (ttl = %s)', self.jid, self.ttl)
        return self.expires_at

    def fail(self, group, message):
        '''Mark the particular job as failed, with the provided type, and a
        more specific message. By `type`, we mean some phrase that might be
        one of several categorical modes of failure. The `message` is
        something more job-specific, like perhaps a traceback.

        This method should __not__ be used to note that a job has been dropped
        or has failed in a transient way. This method __should__ be used to
        note that a job has something really wrong with it that must be
        remedied.

        The motivation behind the `type` is so that similar errors can be
        grouped together. Optionally, updated data can be provided for the job.
        A job in any state can be marked as failed. If it has been given to a
        worker as a job, then its subsequent requests to heartbeat or complete
        that job will fail. Failed jobs are kept until they are canceled or
        completed. __Returns__ the id of the failed job if successful, or
        `False` on failure.'''
        logger.warn('Failing %s (%s): %s', self.jid, group, message)
        return self.client('fail', self.jid, self.client.worker_name, group,
            message, json.dumps(self.data)) or False

    def track(self):
        '''Begin tracking this job'''
        return self.client('track', 'track', self.jid)

    def untrack(self):
        '''Stop tracking this job'''
        return self.client('track', 'untrack', self.jid)

    def retry(self, delay=0):
        '''Retry this job in a little bit, in the same queue. This is meant
        for the times when you detect a transient failure yourself'''
        return self.client('retry', self.jid, self.queue_name,
            self.worker_name, delay)

    def depend(self, *args):
        '''If and only if a job already has other dependencies, this will add
        more jids to the list of this job's dependencies.'''
        return self.client('depends', self.jid, 'on', *args) or False

    def undepend(self, *args, **kwargs):
        '''Remove specific (or all) job dependencies from this job:

            job.remove(jid1, jid2)
            job.remove(all=True)'''
        if kwargs.get('all', False):
            return self.client('depends', self.jid, 'off', 'all') or False
        else:
            return self.client('depends', self.jid, 'off', *args) or False

    def timeout(self):
        '''Time out this job'''
        self.client('timeout', self.jid)


class RecurringJob(BaseJob):
    '''Recurring Job object'''
    def __init__(self, client, **kwargs):
        BaseJob.__init__(self, client, **kwargs)
        for att in ['jid', 'priority', 'tags',
            'retries', 'interval', 'count']:
            object.__setattr__(self, att, kwargs[att])
        object.__setattr__(self, 'client', client)
        object.__setattr__(self, 'klass_name', kwargs['klass'])
        object.__setattr__(self, 'queue_name', kwargs['queue'])
        object.__setattr__(self, 'tags', self.tags or [])
        object.__setattr__(self, 'data', json.loads(kwargs['data']))

    def __setattr__(self, key, value):
        if key in ('priority', 'retries', 'interval'):
            return self.client('recur.update', self.jid, key, value
                ) and object.__setattr__(self, key, value)
        if key == 'data':
            return self.client('recur.update', self.jid, key, json.dumps(value)
                ) and object.__setattr__(self, 'data', value)
        if key == 'klass':
            name = value.__module__ + '.' + value.__name__
            return self.client('recur.update', self.jid, 'klass', name
                ) and object.__setattr__(self, 'klass_name', name
                ) and object.__setattr__(self, 'klass', value)
        return object.__setattr__(self, key, value)

    def __getattr__(self, key):
        if key == 'next':
            # The time (seconds since epoch) until the next time it's run
            return self.client.redis.zscore(
                'ql:q:' + self.queue_name + '-recur', self.jid)
        return BaseJob.__getattr__(self, key)

    def move(self, queue):
        '''Make this recurring job attached to another queue'''
        return self.client('recur.update', self.jid, 'queue', queue)

    def cancel(self):
        '''Cancel all future recurring jobs'''
        self.client('unrecur', self.jid)

    def tag(self, *tags):
        '''Add tags to this recurring job'''
        return self.client('recur.tag', self.jid, *tags)

    def untag(self, *tags):
        '''Remove tags from this job'''
        return self.client('recur.untag', self.jid, *tags)



'''Main qless business'''

import time
import redis
import pkgutil
import logging
import decorator
import simplejson as json

# Internal imports
from .exceptions import QlessException


# Our logger
logger = logging.getLogger('qless')
formatter = logging.Formatter(
    '%(asctime)s | PID %(process)d | [%(levelname)s] %(message)s')
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.FATAL)


def retry(*excepts):
    '''A decorator to specify a bunch of exceptions that should be caught
    and the job retried. It turns out this comes up with relative frequency'''
    @decorator.decorator
    def new_func(func, job):
        '''No docstring'''
        try:
            func(job)
        except tuple(excepts):
            job.retry()
    return new_func


class Jobs(object):
    '''Class for accessing jobs and job information lazily'''
    def __init__(self, client):
        self.client = client

    def complete(self, offset=0, count=25):
        '''Return the paginated jids of complete jobs'''
        return self.client('jobs', 'complete', offset, count)

    def tracked(self):
        '''Return an array of job objects that are being tracked'''
        results = json.loads(self.client('track'))
        results['jobs'] = [Job(self, **job) for job in results['jobs']]
        return results

    def tagged(self, tag, offset=0, count=25):
        '''Return the paginated jids of jobs tagged with a tag'''
        return json.loads(self.client('tag', 'get', tag, offset, count))

    def failed(self, group=None, start=0, limit=25):
        '''If no group is provided, this returns a JSON blob of the counts of
        the various types of failures known. If a type is provided, returns
        paginated job objects affected by that kind of failure.'''
        if not group:
            return json.loads(self.client('failed'))
        else:
            results = json.loads(
                self.client('failed', group, start, limit))
            results['jobs'] = self.get(*results['jobs'])
            return results

    def get(self, *jids):
        '''Return jobs objects for all the jids'''
        if jids:
            return [
                Job(self.client, **j) for j in
                json.loads(self.client('multiget', *jids))]
        return []

    def __getitem__(self, jid):
        '''Get a job object corresponding to that jid, or ``None`` if it
        doesn't exist'''
        results = self.client('get', jid)
        if not results:
            results = self.client('recur.get', jid)
            if not results:
                return None
            return RecurringJob(self.client, **json.loads(results))
        return Job(self.client, **json.loads(results))


class Workers(object):
    '''Class for accessing worker information lazily'''
    def __init__(self, clnt):
        self.client = clnt

    def __getattr__(self, attr):
        '''What workers are workers, and how many jobs are they running'''
        if attr == 'counts':
            return json.loads(self.client('workers'))
        raise AttributeError('qless.Workers has no attribute %s' % attr)

    def __getitem__(self, worker_name):
        '''Which jobs does a particular worker have running'''
        result = json.loads(
            self.client('workers', worker_name))
        result['jobs']    = result['jobs'] or []
        result['stalled'] = result['stalled'] or []
        return result


class Queues(object):
    '''Class for accessing queues lazily'''
    def __init__(self, clnt):
        self.client = clnt

    def __getattr__(self, attr):
        '''What queues are there, and how many jobs do they have running,
        waiting, scheduled, etc.'''
        if attr == 'counts':
            return json.loads(self.client('queues'))
        raise AttributeError('qless.Queues has no attribute %s' % attr)

    def __getitem__(self, queue_name):
        '''Get a queue object associated with the provided queue name'''
        return Queue(queue_name, self.client, self.client.worker_name)


class Client(object):
    '''Basic qless client object.'''
    def __init__(self, url='redis://localhost:6379', hostname=None, **kwargs):
        import socket
        # This is our unique idenitifier as a worker
        self.worker_name = hostname or socket.gethostname()
        # This is just the redis instance we're connected to conceivably
        # someone might want to work with multiple instances simultaneously.
        self.redis = redis.Redis.from_url(url, **kwargs)
        self.jobs = Jobs(self)
        self.queues = Queues(self)
        self.config = Config(self)
        self.workers = Workers(self)

        # We now have a single unified core script.
        data = pkgutil.get_data('qless', 'qless-core/qless.lua')
        self._lua = self.redis.register_script(data)

    def __getattr__(self, key):
        if key == 'events':
            self.events = Events(self.redis)
            return self.events
        raise AttributeError('%s has no attribute %s' % (
            self.__class__.__module__ + '.' + self.__class__.__name__, key))

    def __call__(self, command, *args):
        lua_args = [command, repr(time.time())]
        lua_args.extend(args)
        try:
            return self._lua(keys=[], args=lua_args)
        except redis.ResponseError as exc:
            raise QlessException(str(exc))

    def track(self, jid):
        '''Begin tracking this job'''
        return self('track', 'track', jid)

    def untrack(self, jid):
        '''Stop tracking this job'''
        return self('track', 'untrack', jid)

    def tags(self, offset=0, count=100):
        '''The most common tags among jobs'''
        return json.loads(self('tag', 'top', offset, count))

    def unfail(self, group, queue, count=500):
        '''Move jobs from the failed group to the provided queue'''
        return self('unfail', queue, group, count)

from .job import Job, RecurringJob
from .queue import Queue
from .config import Config
from .listener import Events



'''Tests about the config class'''

from common import TestQless


class TestConfig(TestQless):
    '''Test the config class'''
    def test_set_get_unset(self):
        '''Basic set/get/unset'''
        self.assertEqual(self.client.config['foo'], None)
        self.client.config['foo'] = 5
        self.assertEqual(self.client.config['foo'], 5)
        del self.client.config['foo']
        self.assertEqual(self.client.config['foo'], None)

    def test_get_all(self):
        '''Ensure we can get all the configuration'''
        self.assertEqual(self.client.config.all, {
            'application': 'qless',
            'grace-period': 10,
            'heartbeat': 60,
            'histogram-history': 7,
            'jobs-history': 604800,
            'jobs-history-count': 50000,
            'stats-history': 30
        })

    def test_clear(self):
        '''Can unset all keys'''
        original = dict(self.client.config.items())
        for key in self.client.config.keys():
            self.client.config[key] = 1
        for value in self.client.config.values():
            self.assertEqual(value, '1')
        self.client.config.clear()
        self.assertEqual(self.client.config.all, original)

    def test_attribute_error(self):
        '''Only has the 'all' attribute'''
        self.assertRaises(AttributeError, lambda: self.client.config.foo)

    def test_len(self):
        '''We can see how many items are in the config'''
        self.assertEqual(len(self.client.config), 7)

    def test_contains(self):
        '''We can use the 'in' syntax'''
        self.assertFalse('foo' in self.client.config)
        self.client.config['foo'] = 5
        self.assertTrue('foo' in self.client.config)

    def test_iter(self):
        '''We can iterate over the config'''
        self.assertEqual(
            [key for key in self.client.config], self.client.config.keys())

    def test_get(self):
        '''We can use dictionary-style get'''
        self.assertFalse('foo' in self.client.config)
        self.assertEqual(self.client.config.get('foo', 5), 5)

    def test_pop(self):
        '''We can use dictionary-style pop'''
        self.assertFalse('foo' in self.client.config)
        self.client.config['foo'] = 5
        self.assertEqual(self.client.config.pop('foo'), 5)
        self.assertFalse('foo' in self.client.config)

    def test_update(self):
        '''We can use dictionary-style update'''
        updated = dict((key, '1') for key in self.client.config)
        self.assertNotEqual(self.client.config.all, updated)
        self.client.config.update(updated)
        self.assertEqual(self.client.config.all, updated)



'''Basic tests about the Job class'''

from common import TestQless
from qless.job import Job, BaseJob


class Foo(object):
    '''A dummy job'''
    @staticmethod
    def bar(job):
        '''A dummy method'''
        job['foo'] = 'bar'
        job.complete()

    def nonstatic(self, job):
        '''A nonstatic method'''
        pass


class TestJob(TestQless):
    '''Test the Job class'''
    def test_attributes(self):
        '''Has all the basic attributes we'd expect'''
        self.client.queues['foo'].put('Foo', {'whiz': 'bang'}, jid='jid',
            tags=['foo'], retries=3)
        job = self.client.jobs['jid']
        atts = ['data', 'jid', 'priority', 'klass_name', 'queue_name', 'tags',
            'expires_at', 'original_retries', 'retries_left', 'worker_name',
            'dependents', 'dependencies']
        values = [getattr(job, att) for att in atts]
        self.assertEqual(dict(zip(atts, values)), {
            'data': {'whiz': 'bang'},
            'dependencies': [],
            'dependents': [],
            'expires_at': 0,
            'jid': 'jid',
            'klass_name': 'Foo',
            'original_retries': 3,
            'priority': 0,
            'queue_name': 'foo',
            'retries_left': 3,
            'tags': ['foo'],
            'worker_name': u''
        })

    def test_set_priority(self):
        '''We can set a job's priority'''
        self.client.queues['foo'].put('Foo', {}, jid='jid', priority=0)
        self.assertEqual(self.client.jobs['jid'].priority, 0)
        self.client.jobs['jid'].priority = 10
        self.assertEqual(self.client.jobs['jid'].priority, 10)

    def test_queue(self):
        '''Exposes a queue object'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertEqual(self.client.jobs['jid'].queue.name, 'foo')

    def test_klass(self):
        '''Exposes the class for a job'''
        self.client.queues['foo'].put(Job, {}, jid='jid')
        self.assertEqual(self.client.jobs['jid'].klass, Job)

    def test_ttl(self):
        '''Exposes the ttl for a job'''
        self.client.config['heartbeat'] = 10
        self.client.queues['foo'].put(Job, {}, jid='jid')
        self.client.queues['foo'].pop()
        self.assertTrue(self.client.jobs['jid'].ttl < 10)
        self.assertTrue(self.client.jobs['jid'].ttl > 9)

    def test_attribute_error(self):
        '''Raises an attribute error for nonexistent attributes'''
        self.client.queues['foo'].put(Job, {}, jid='jid')
        self.assertRaises(AttributeError, lambda: self.client.jobs['jid'].foo)

    def test_cancel(self):
        '''Exposes the cancel method'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.jobs['jid'].cancel()
        self.assertEqual(self.client.jobs['jid'], None)

    def test_tag_untag(self):
        '''Exposes a way to tag and untag a job'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.jobs['jid'].tag('foo')
        self.assertEqual(self.client.jobs['jid'].tags, ['foo'])
        self.client.jobs['jid'].untag('foo')
        self.assertEqual(self.client.jobs['jid'].tags, [])

    def test_getitem(self):
        '''Exposes job data through []'''
        self.client.queues['foo'].put('Foo', {'foo': 'bar'}, jid='jid')
        self.assertEqual(self.client.jobs['jid']['foo'], 'bar')

    def test_setitem(self):
        '''Sets jobs data through []'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        job = self.client.jobs['jid']
        job['foo'] = 'bar'
        self.assertEqual(job['foo'], 'bar')

    def test_move(self):
        '''Able to move jobs through the move method'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.jobs['jid'].move('bar')
        self.assertEqual(self.client.jobs['jid'].queue.name, 'bar')

    def test_complete(self):
        '''Able to complete a job'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.queues['foo'].pop().complete()
        self.assertEqual(self.client.jobs['jid'].state, 'complete')

    def test_advance(self):
        '''Able to advance a job to another queue'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.queues['foo'].pop().complete('bar')
        self.assertEqual(self.client.jobs['jid'].state, 'waiting')

    def test_heartbeat(self):
        '''Provides access to heartbeat'''
        self.client.config['heartbeat'] = 10
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        job = self.client.queues['foo'].pop()
        before = job.ttl
        self.client.config['heartbeat'] = 20
        job.heartbeat()
        self.assertTrue(job.ttl > before)

    def test_heartbeat_fail(self):
        '''Failed heartbeats raise an error'''
        from qless.exceptions import LostLockException
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertRaises(LostLockException, self.client.jobs['jid'].heartbeat)

    def test_track_untrack(self):
        '''Exposes a track, untrack method'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertFalse(self.client.jobs['jid'].tracked)
        self.client.jobs['jid'].track()
        self.assertTrue(self.client.jobs['jid'].tracked)
        self.client.jobs['jid'].untrack()
        self.assertFalse(self.client.jobs['jid'].tracked)

    def test_depend_undepend(self):
        '''Exposes a depend, undepend methods'''
        self.client.queues['foo'].put('Foo', {}, jid='a')
        self.client.queues['foo'].put('Foo', {}, jid='b')
        self.client.queues['foo'].put('Foo', {}, jid='c', depends=['a'])
        self.assertEqual(self.client.jobs['c'].dependencies, ['a'])
        self.client.jobs['c'].depend('b')
        self.assertEqual(self.client.jobs['c'].dependencies, ['a', 'b'])
        self.client.jobs['c'].undepend('a')
        self.assertEqual(self.client.jobs['c'].dependencies, ['b'])
        self.client.jobs['c'].undepend(all=True)
        self.assertEqual(self.client.jobs['c'].dependencies, [])

    def test_retry_fail(self):
        '''Retry raises an error if retry fails'''
        from qless.exceptions import QlessException
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertRaises(QlessException, self.client.jobs['jid'].retry)

    def test_repr(self):
        '''Has a reasonable repr'''
        self.client.queues['foo'].put(Job, {}, jid='jid')
        self.assertEqual(repr(self.client.jobs['jid']), '<qless.job.Job jid>')

    def test_no_method(self):
        '''Raises an error if the class doesn't have the method'''
        self.client.queues['foo'].put(Foo, {}, jid='jid')
        self.client.queues['foo'].pop().process()
        job = self.client.jobs['jid']
        self.assertEqual(job.state, 'failed')
        self.assertEqual(job.failure['group'], 'foo-method-missing')

    def test_no_import(self):
        '''Raises an error if it can't import the class'''
        self.client.queues['foo'].put('foo.Foo', {}, jid='jid')
        self.client.queues['foo'].pop().process()
        job = self.client.jobs['jid']
        self.assertEqual(job.state, 'failed')
        self.assertEqual(job.failure['group'], 'foo-ImportError')

    def test_nonstatic(self):
        '''Rasises an error if the relevant function's not static'''
        self.client.queues['nonstatic'].put(Foo, {}, jid='jid')
        self.client.queues['nonstatic'].pop().process()
        job = self.client.jobs['jid']
        self.assertEqual(job.state, 'failed')
        self.assertEqual(job.failure['group'], 'nonstatic-method-type')

    def test_reload(self):
        '''Ensure that nothing blows up if we reload a class'''
        self.client.queues['foo'].put(Foo, {}, jid='jid')
        self.assertEqual(self.client.jobs['jid'].klass, Foo)
        Job.reload(self.client.jobs['jid'].klass_name)
        self.assertEqual(self.client.jobs['jid'].klass, Foo)


class TestRecurring(TestQless):
    def test_attributes(self):
        '''We can access all the recurring attributes'''
        self.client.queues['foo'].recur('Foo', {'whiz': 'bang'}, 60, jid='jid',
            tags=['foo'], retries=3)
        job = self.client.jobs['jid']
        atts = ['data', 'jid', 'priority', 'klass_name', 'queue_name', 'tags',
            'retries', 'interval', 'count']
        values = [getattr(job, att) for att in atts]
        self.assertEqual(dict(zip(atts, values)), {
            'count': 0,
            'data': {'whiz': 'bang'},
            'interval': 60,
            'jid': 'jid',
            'klass_name': 'Foo',
            'priority': 0,
            'queue_name': 'foo',
            'retries': 3,
            'tags': ['foo']
        })

    def test_set_priority(self):
        '''We can set priority on recurring jobs'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid', priority=0)
        self.client.jobs['jid'].priority = 10
        self.assertEqual(self.client.jobs['jid'].priority, 10)

    def test_set_retries(self):
        '''We can set retries'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid', retries=2)
        self.client.jobs['jid'].retries = 2
        self.assertEqual(self.client.jobs['jid'].retries, 2)

    def test_set_interval(self):
        '''We can set the interval'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].interval = 10
        self.assertEqual(self.client.jobs['jid'].interval, 10)

    def test_set_data(self):
        '''We can set the job data'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].data = {'foo': 'bar'}
        self.assertEqual(self.client.jobs['jid'].data, {'foo': 'bar'})

    def test_set_klass(self):
        '''We can set the klass'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].klass = Foo
        self.assertEqual(self.client.jobs['jid'].klass, Foo)

    def test_get_next(self):
        '''Exposes the next time a job will run'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        nxt = self.client.jobs['jid'].next
        self.client.queues['foo'].pop()
        self.assertTrue(abs(self.client.jobs['jid'].next - nxt - 60) < 1)

    def test_attribute_error(self):
        '''Raises attribute errors for non-attributes'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.assertRaises(AttributeError, lambda: self.client.jobs['jid'].foo)

    def test_move(self):
        '''Exposes a way to move a job'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].move('bar')
        self.assertEqual(self.client.jobs['jid'].queue.name, 'bar')

    def test_cancel(self):
        '''Exposes a way to cancel jobs'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].cancel()
        self.assertEqual(self.client.jobs['jid'], None)

    def test_tag_untag(self):
        '''Exposes a way to tag jobs'''
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.client.jobs['jid'].tag('foo')
        self.assertEqual(self.client.jobs['jid'].tags, ['foo'])
        self.client.jobs['jid'].untag('foo')
        self.assertEqual(self.client.jobs['jid'].tags, [])



'''Some utilities for profiling'''

import redis
from collections import defaultdict


class Profiler(object):
    '''Profiling a series of requests. Initialized with a Qless client'''
    @staticmethod
    def clone(client):
        '''Clone the redis client to be slowlog-compatible'''
        kwargs = client.redis.connection_pool.connection_kwargs
        kwargs['parser_class'] = redis.connection.PythonParser
        pool = redis.connection.ConnectionPool(**kwargs)
        return redis.Redis(connection_pool=pool)

    @staticmethod
    def pretty(timings, label):
        '''Print timing stats'''
        results = [(sum(values), len(values), key)
            for key, values in timings.items()]
        print label
        print '=' * 65
        print '%20s => %13s | %8s | %13s' % (
            'Command', 'Average', '# Calls', 'Total time')
        print '-' * 65
        for total, length, key in sorted(results, reverse=True):
            print '%20s => %10.5f us | %8i | %10i us' % (
                key, float(total) / length, length, total)

    def __init__(self, client):
        self._client = self.clone(client)
        self._configs = None
        self._timings = defaultdict(list)
        self._commands = {}

    def start(self):
        '''Get ready for a profiling run'''
        self._configs = self._client.config_get('slow-*')
        self._client.config_set('slowlog-max-len', 100000)
        self._client.config_set('slowlog-log-slower-than', 0)
        self._client.execute_command('slowlog', 'reset')

    def stop(self):
        '''Set everything back to normal and collect our data'''
        for key, value in self._configs.items():
            self._client.config_set(key, value)
        logs = self._client.execute_command('slowlog', 'get', 100000)
        current = {
            'name': None, 'accumulated': defaultdict(list)
        }
        for _, _, duration, request in logs:
            command = request[0]
            if command == 'slowlog':
                continue
            if 'eval' in command.lower():
                subcommand = request[3]
                self._timings['qless-%s' % subcommand].append(duration)
                if current['name']:
                    if current['name'] not in self._commands:
                        self._commands[current['name']] = defaultdict(list)
                    for key, values in current['accumulated'].items():
                        self._commands[current['name']][key].extend(values)
                current = {
                    'name': subcommand, 'accumulated': defaultdict(list)
                }
            else:
                self._timings[command].append(duration)
                if current['name']:
                    current['accumulated'][command].append(duration)
        # Include the last
        if current['name']:
            if current['name'] not in self._commands:
                self._commands[current['name']] = defaultdict(list)
            for key, values in current['accumulated'].items():
                self._commands[current['name']][key].extend(values)

    def display(self):
        '''Print the results of this profiling'''
        self.pretty(self._timings, 'Raw Redis Commands')
        print ''
        for key, value in self._commands.items():
            self.pretty(value, 'Qless "%s" Command' % key)
            print ''

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, typ, value, trace):
        self.stop()
        self.display()
        if typ:
            raise typ, value, trace



'''Test the forking worker'''

# Internal imports
from common import TestQless

import os
import time
import signal
import threading

# The stuff we're actually testing
import qless
from qless.workers import Worker
from qless.workers.forking import ForkingWorker


class Foo(object):
    '''Dummy class'''
    @staticmethod
    def foo(job):
        '''Fall on your sword!'''
        os.kill(os.getpid(), signal.SIGKILL)


class CWD(object):
    '''Completes with our current working directory'''
    @staticmethod
    def foo(job):
        '''Puts your current working directory in the job data'''
        job.data['cwd'] = os.getcwd()
        job.complete()
        os.kill(os.getpid(), signal.SIGKILL)


class PatchedForkingWorker(ForkingWorker):
    '''A forking worker that doesn't register signal handlers'''
    def signals(self, signals=()):
        '''Do not actually register signal handlers'''
        pass


class TestWorker(TestQless):
    '''Test the worker'''
    def setUp(self):
        TestQless.setUp(self)
        self.worker = PatchedForkingWorker(
            ['foo'], self.client, workers=1, interval=1)
        self.queue = self.client.queues['foo']
        self.thread = None

    def tearDown(self):
        if self.thread:
            self.thread.join()
        TestQless.tearDown(self)

    def test_respawn(self):
        '''It respawns workers as needed'''
        self.thread = threading.Thread(target=self.worker.run)
        self.thread.start()
        time.sleep(0.1)
        self.worker.shutdown = True
        self.queue.put(Foo, {})
        self.thread.join(1)
        self.assertFalse(self.thread.is_alive())

    def test_cwd(self):
        '''Should set the child's cwd appropriately'''
        self.thread = threading.Thread(target=self.worker.run)
        self.thread.start()
        time.sleep(0.1)
        self.worker.shutdown = True
        jid = self.queue.put(CWD, {})
        self.thread.join(1)
        self.assertFalse(self.thread.is_alive())
        expected = os.path.join(os.getcwd(), 'qless-py-workers/sandbox-0')
        self.assertEqual(self.client.jobs[jid]['cwd'], expected)

    def test_spawn_klass_string(self):
        '''Should be able to import by class string'''
        worker = PatchedForkingWorker(['foo'], self.client)
        worker.klass = 'qless.workers.serial.SerialWorker'
        self.assertIsInstance(worker.spawn(), Worker)

    def test_spawn(self):
        '''It gives us back a worker instance'''
        self.assertIsInstance(self.worker.spawn(), Worker)



'''A worker that forks child processes'''

import os
import psutil
import signal

# Internal imports
from . import Worker
from qless import logger, util
from .serial import SerialWorker


class ForkingWorker(Worker):
    '''A worker that forks child processes'''
    def __init__(self, *args, **kwargs):
        Worker.__init__(self, *args, **kwargs)
        # Worker class to use
        self.klass = self.kwargs.pop('klass', SerialWorker)
        # How many children to launch
        self.count = self.kwargs.pop('workers', 0) or psutil.NUM_CPUS
        # A dictionary of child pids to information about them
        self.sandboxes = {}
        # Whether or not we're supposed to shutdown
        self.shutdown = False

    def stop(self, sig=signal.SIGINT):
        '''Stop all the workers, and then wait for them'''
        for cpid in self.sandboxes.keys():
            logger.warn('Stopping %i...' % cpid)
            try:
                os.kill(cpid, sig)
            except OSError:  # pragma: no cover
                logger.exception('Error stopping %s...' % cpid)

        # While we still have children running, wait for them
        for cpid in self.sandboxes.keys():
            try:
                logger.info('Waiting for %i...' % cpid)
                pid, status = os.waitpid(cpid, 0)
                logger.warn('%i stopped with status %i' % (pid, status >> 8))
            except OSError:  # pragma: no cover
                logger.exception('Error waiting for %i...' % cpid)
            finally:
                self.sandboxes.pop(cpid, None)

    def spawn(self, **kwargs):
        '''Return a new worker for a child process'''
        copy = dict(self.kwargs)
        copy.update(kwargs)
        # Apparently there's an issue with importing gevent in the parent
        # process and then using it int he child. This is meant to relieve that
        # problem by allowing `klass` to be specified as a string.
        if isinstance(self.klass, basestring):
            self.klass = util.import_class(self.klass)
        return self.klass(self.queues, self.client, **copy)

    def run(self):
        '''Run this worker'''
        self.signals(('TERM', 'INT', 'QUIT'))
        # Divide up the jobs that we have to divy up between the workers. This
        # produces evenly-sized groups of jobs
        resume = self.divide(self.resume, self.count)
        for index in range(self.count):
            # The sandbox for the child worker
            sandbox = os.path.join(
                os.getcwd(), 'qless-py-workers', 'sandbox-%s' % index)
            cpid = os.fork()
            if cpid:
                logger.info('Spawned worker %i' % cpid)
                self.sandboxes[cpid] = sandbox
            else:  # pragma: no cover
                # Move to the sandbox as the current working directory
                with Worker.sandbox(sandbox):
                    os.chdir(sandbox)
                    self.spawn(resume=resume[index], sandbox=sandbox).run()
                    exit(0)

        try:
            while not self.shutdown:
                pid, status = os.wait()
                logger.warn('Worker %i died with status %i from signal %i' % (
                    pid, status >> 8, status & 0xff))
                sandbox = self.sandboxes.pop(pid)
                cpid = os.fork()
                if cpid:
                    logger.info('Spawned replacement worker %i' % cpid)
                    self.sandboxes[cpid] = sandbox
                else:  # pragma: no cover
                    with Worker.sandbox(sandbox):
                        os.chdir(sandbox)
                        self.spawn(sandbox=sandbox).run()
                        exit(0)
        finally:
            self.stop(signal.SIGKILL)

    def handler(self, signum, frame):  # pragma: no cover
        '''Signal handler for this process'''
        if signum in (signal.SIGTERM, signal.SIGINT, signal.SIGQUIT):
            for cpid in self.sandboxes.keys():
                try:
                    os.kill(cpid, signum)
                except OSError:  # pragma: no cover
                    logger.exception(
                        'Failed to send %s to %s...' % (signum, cpid))
            exit(0)



#! /usr/bin/env python

'''Our Queue and supporting classes'''

import time
import uuid
from qless.job import Job
import simplejson as json


class Jobs(object):
    '''A proxy object for queue-specific job information'''
    def __init__(self, name, client):
        self.name = name
        self.client = client

    def running(self, offset=0, count=25):
        '''Return all the currently-running jobs'''
        return self.client('jobs', 'running', self.name, offset, count)

    def stalled(self, offset=0, count=25):
        '''Return all the currently-stalled jobs'''
        return self.client('jobs', 'stalled', self.name, offset, count)

    def scheduled(self, offset=0, count=25):
        '''Return all the currently-scheduled jobs'''
        return self.client('jobs', 'scheduled', self.name, offset, count)

    def depends(self, offset=0, count=25):
        '''Return all the currently dependent jobs'''
        return self.client('jobs', 'depends', self.name, offset, count)

    def recurring(self, offset=0, count=25):
        '''Return all the recurring jobs'''
        return self.client('jobs', 'recurring', self.name, offset, count)


class Queue(object):
    '''The Queue class'''
    def __init__(self, name, client, worker_name):
        self.name        = name
        self.client      = client
        self.worker_name = worker_name
        self._hb         = 60

    def __getattr__(self, key):
        if key == 'jobs':
            self.jobs = Jobs(self.name, self.client)
            return self.jobs
        if key == 'counts':
            return json.loads(self.client('queues', self.name))
        if key == 'heartbeat':
            conf = self.client.config.all
            return int(conf.get(
                self.name + '-heartbeat', conf.get('heartbeat', 60)))
        raise AttributeError('qless.Queue has no attribute %s' % key)

    def __setattr__(self, key, value):
        if key == 'heartbeat':
            self.client.config[self.name + '-heartbeat'] = value
        else:
            object.__setattr__(self, key, value)

    def class_string(self, klass):
        '''Return a string representative of the class'''
        if isinstance(klass, basestring):
            return klass
        return klass.__module__ + '.' + klass.__name__

    def put(self, klass, data, priority=None, tags=None, delay=None,
        retries=None, jid=None, depends=None):
        '''Either create a new job in the provided queue with the provided
        attributes, or move that job into that queue. If the job is being
        serviced by a worker, subsequent attempts by that worker to either
        `heartbeat` or `complete` the job should fail and return `false`.

        The `priority` argument should be negative to be run sooner rather
        than later, and positive if it's less important. The `tags` argument
        should be a JSON array of the tags associated with the instance and
        the `valid after` argument should be in how many seconds the instance
        should be considered actionable.'''
        return self.client('put', self.name,
            jid or uuid.uuid4().hex,
            self.class_string(klass),
            json.dumps(data),
            delay or 0,
            'priority', priority or 0,
            'tags', json.dumps(tags or []),
            'retries', retries or 5,
            'depends', json.dumps(depends or [])
        )

    def recur(self, klass, data, interval, offset=0, priority=None, tags=None,
        retries=None, jid=None):
        '''Place a recurring job in this queue'''
        return self.client('recur', self.name,
            jid or uuid.uuid4().hex,
            self.class_string(klass),
            json.dumps(data),
            'interval', interval, offset,
            'priority', priority or 0,
            'tags', json.dumps(tags or []),
            'retries', retries or 5
        )

    def pop(self, count=None):
        '''Passing in the queue from which to pull items, the current time,
        when the locks for these returned items should expire, and the number
        of items to be popped off.'''
        results = [Job(self.client, **job) for job in json.loads(
            self.client('pop', self.name, self.worker_name, count or 1))]
        if count == None:
            return (len(results) and results[0]) or None
        return results

    def peek(self, count=None):
        '''Similar to the pop command, except that it merely peeks at the next
        items'''
        results = [Job(self.client, **rec) for rec in json.loads(
            self.client('peek', self.name, count or 1))]
        if count == None:
            return (len(results) and results[0]) or None
        return results

    def stats(self, date=None):
        '''Return the current statistics for a given queue on a given date.
        The results are returned are a JSON blob::

            {
                'total'    : ...,
                'mean'     : ...,
                'variance' : ...,
                'histogram': [
                    ...
                ]
            }

        The histogram's data points are at the second resolution for the first
        minute, the minute resolution for the first hour, the 15-minute
        resolution for the first day, the hour resolution for the first 3
        days, and then at the day resolution from there on out. The
        `histogram` key is a list of those values.'''
        return json.loads(
            self.client('stats', self.name, date or repr(time.time())))

    def __len__(self):
        return self.client('length', self.name)



'''Some utility functions'''


def import_class(klass):
    '''Import the named class and return that class'''
    mod = __import__(klass.rpartition('.')[0])
    for segment in klass.split('.')[1:-1]:
        mod = getattr(mod, segment)
    return getattr(mod, klass.rpartition('.')[2])



'''Basic tests about the client'''

from common import TestQless


class TestClient(TestQless):
    '''Test the client'''
    def test_track(self):
        '''Gives us access to track and untrack jobs'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.track('jid')
        self.assertEqual(self.client.jobs.tracked()['jobs'][0].jid, 'jid')
        self.client.untrack('jid')
        self.assertEqual(self.client.jobs.tracked(),
            {'jobs': [], 'expired': {}})

    def test_attribute_error(self):
        '''Throws AttributeError for non-attributes'''
        self.assertRaises(AttributeError, lambda: self.client.foo)

    def test_tags(self):
        '''Provides access to top tags'''
        self.assertEqual(self.client.tags(), {})
        for _ in range(10):
            self.client.queues['foo'].put('Foo', {}, tags=['foo'])
        self.assertEqual(self.client.tags(), ['foo'])

    def test_unfail(self):
        '''Provides access to unfail'''
        jids = map(str, range(10))
        for jid in jids:
            print 'Popping %s' % jid
            self.client.queues['foo'].put('Foo', {}, jid=jid)
            self.client.queues['foo'].pop().fail('foo', 'bar')
        for jid in jids:
            self.assertEqual(self.client.jobs[jid].state, 'failed')
        self.client.unfail('foo', 'foo')
        for jid in jids:
            self.assertEqual(self.client.jobs[jid].state, 'waiting')


class TestJobs(TestQless):
    '''Test the Jobs class'''
    def test_basic(self):
        '''Can give us access to jobs'''
        self.assertEqual(self.client.jobs['jid'], None)
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertNotEqual(self.client.jobs['jid'], None)

    def test_recurring(self):
        '''Can give us access to recurring jobs'''
        self.assertEqual(self.client.jobs['jid'], None)
        self.client.queues['foo'].recur('Foo', {}, 60, jid='jid')
        self.assertNotEqual(self.client.jobs['jid'], None)

    def test_complete(self):
        '''Can give us access to complete jobs'''
        self.assertEqual(self.client.jobs.complete(), [])
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.queues['foo'].pop().complete()
        self.assertEqual(self.client.jobs.complete(), ['jid'])

    def test_tracked(self):
        '''Gives us access to tracked jobs'''
        self.assertEqual(self.client.jobs.tracked(),
            {'jobs': [], 'expired': {}})
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.track('jid')
        self.assertEqual(self.client.jobs.tracked()['jobs'][0].jid, 'jid')

    def test_tagged(self):
        '''Gives us access to tagged jobs'''
        self.assertEqual(self.client.jobs.tagged('foo'),
            {'total': 0, 'jobs': {}})
        self.client.queues['foo'].put('Foo', {}, jid='jid', tags=['foo'])
        self.assertEqual(self.client.jobs.tagged('foo')['jobs'][0], 'jid')

    def test_failed(self):
        '''Gives us access to failed jobs'''
        self.assertEqual(self.client.jobs.failed('foo'),
            {'total': 0, 'jobs': []})
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.queues['foo'].pop().fail('foo', 'bar')
        self.assertEqual(self.client.jobs.failed('foo')['jobs'][0].jid, 'jid')

    def test_failures(self):
        '''Gives us access to failure types'''
        self.assertEqual(self.client.jobs.failed(), {})
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.queues['foo'].pop().fail('foo', 'bar')
        self.assertEqual(self.client.jobs.failed(), {'foo': 1})


class TestQueues(TestQless):
    '''Test the Queues class'''
    def test_basic(self):
        '''Gives us access to queues'''
        self.assertNotEqual(self.client.queues['foo'], None)

    def test_counts(self):
        '''Gives us access to counts'''
        self.assertEqual(self.client.queues.counts, {})
        self.client.queues['foo'].put('Foo', {})
        self.assertEqual(self.client.queues.counts, [{
            'scheduled': 0,
            'name': 'foo',
            'paused': False,
            'waiting': 1,
            'depends': 0,
            'running': 0,
            'stalled': 0,
            'recurring': 0
        }])

    def test_attribute_error(self):
        '''Raises AttributeErrors for non-attributes'''
        self.assertRaises(AttributeError, lambda: self.client.queues.foo)


class TestWorkers(TestQless):
    '''Test the Workers class'''
    def test_individual(self):
        '''Gives us access to individual workers'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertEqual(self.client.workers['worker'],
            {'jobs': [], 'stalled': []})
        self.worker.queues['foo'].pop()
        self.assertEqual(self.client.workers['worker'],
            {'jobs': ['jid'], 'stalled': []})

    def test_counts(self):
        '''Gives us access to worker counts'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertEqual(self.client.workers.counts, {})
        self.worker.queues['foo'].pop()
        self.assertEqual(self.client.workers.counts,
            [{'jobs': 1, 'name': 'worker', 'stalled': 0}])

    def test_attribute_error(self):
        '''Raises AttributeErrors for non-attributes'''
        self.assertRaises(AttributeError, lambda: self.client.workers.foo)


# This is used for TestRetry
class Foo(object):
    from qless import retry

    @staticmethod
    @retry(ValueError)
    def process(job):
        '''This is supposed to raise an Exception'''
        if 'valueerror' in job.tags:
            print 'Raising ValueError'
            raise ValueError('Foo')
        else:
            print 'Raising Exception'
            raise Exception('Foo')


class TestRetry(TestQless):
    '''Test the retry decorator'''
    def test_basic(self):
        '''Ensure the retry decorator works'''
        # The first time, it should just be retries automatically
        self.client.queues['foo'].put(Foo, {}, tags=['valueerror'], jid='jid')
        self.client.queues['foo'].pop().process()
        # Not remove the tag so it should fail
        self.client.jobs['jid'].untag('valueerror')
        self.client.queues['foo'].pop().process()
        self.assertEqual(self.client.jobs['jid'].state, 'failed')

    def test_docstring(self):
        '''Retry decorator should preserve docstring'''
        self.assertEqual(Foo.process.__doc__,
            'This is supposed to raise an Exception')



'''A class that listens to pubsub channels and can unlisten'''

import logging
import threading
import contextlib

# Our logger
logger = logging.getLogger('qless')


class Listener(object):
    '''A class that listens to pubsub channels and can unlisten'''
    def __init__(self, redis, channels):
        self._pubsub = redis.pubsub()
        self._channels = channels

    def listen(self):
        '''Listen for events as they come in'''
        try:
            self._pubsub.subscribe(self._channels)
            for message in self._pubsub.listen():
                if message['type'] == 'message':
                    yield message
        finally:
            self._channels = []

    def unlisten(self):
        '''Stop listening for events'''
        self._pubsub.unsubscribe(self._channels)

    @contextlib.contextmanager
    def thread(self):
        '''Run in a thread'''
        thread = threading.Thread(target=self.listen)
        thread.start()
        try:
            yield self
        finally:
            self.unlisten()
            thread.join()


class Events(Listener):
    '''A class for handling qless events'''
    namespace = 'ql:'
    events = (
        'canceled', 'completed', 'failed', 'popped',
        'stalled', 'put', 'track', 'untrack'
    )

    def __init__(self, redis):
        Listener.__init__(
            self, redis, [self.namespace + event for event in self.events])
        self._callbacks = dict((k, None) for k in (self.events))

    def listen(self):
        '''Listen for events'''
        for message in Listener.listen(self):
            logger.debug('Message: %s', message)
            # Strip off the 'namespace' from the channel
            channel = message['channel'][len(self.namespace):]
            func = self._callbacks.get(channel)
            if func:
                func(message['data'])

    def on(self, evt, func):
        '''Set a callback handler for a pubsub event'''
        if evt not in self._callbacks:
            raise NotImplementedError('callback "%s"' % evt)
        else:
            self._callbacks[evt] = func

    def off(self, evt):
        '''Deactivate the callback for a pubsub event'''
        return self._callbacks.pop(evt, None)



#! /usr/bin/env python

from distutils.core import setup

setup(
    name                 = 'qless-py',
    version              = '0.10.2',
    description          = 'Redis-based Queue Management',
    long_description     = '''
Redis-based queue management, with heartbeating, job tracking,
stats, notifications, and a whole lot more.''',
    url                  = 'http://github.com/seomoz/qless-py',
    author               = 'Dan Lecocq',
    author_email         = 'dan@seomoz.org',
    license              = "MIT License",
    keywords             = 'redis, qless, job',
    packages             = ['qless', 'qless.workers'],
    package_dir          = {
        'qless': 'qless',
        'qless.workers': 'qless/workers'},
    package_data         = {'qless': ['qless-core/*.lua']},
    include_package_data = True,
    scripts              = ['bin/qless-py-worker'],
    extras_require       = {
        'ps': ['setproctitle']
    },
    install_requires     = [
        'argparse', 'decorator', 'hiredis', 'redis', 'psutil', 'simplejson'],
    classifiers          = [
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ]
)



'''Basic tests about the Job class'''

from common import TestQless


class TestQueue(TestQless):
    '''Test the Job class'''
    def test_jobs(self):
        '''The queue.Jobs class provides access to job counts'''
        queue = self.client.queues['foo']
        queue.put('Foo', {})
        self.assertEqual(queue.jobs.depends(), [])
        self.assertEqual(queue.jobs.running(), [])
        self.assertEqual(queue.jobs.stalled(), [])
        self.assertEqual(queue.jobs.scheduled(), [])
        self.assertEqual(queue.jobs.recurring(), [])

    def test_counts(self):
        '''Provides access to job counts'''
        self.client.queues['foo'].put('Foo', {})
        self.assertEqual(self.client.queues['foo'].counts, {
            'depends': 0,
            'name': 'foo',
            'paused': False,
            'recurring': 0,
            'running': 0,
            'scheduled': 0,
            'stalled': 0,
            'waiting': 1
        })

    def test_heartbeat(self):
        '''Provided access to heartbeat configuration'''
        original = self.client.queues['foo'].heartbeat
        self.client.queues['foo'].heartbeat = 10
        self.assertNotEqual(original, self.client.queues['foo'].heartbeat)

    def test_attribute_error(self):
        '''Raises an attribute error if there is no attribute'''
        self.assertRaises(AttributeError, lambda: self.client.queues['foo'].foo)

    def test_multipop(self):
        '''Exposes multi-pop'''
        self.client.queues['foo'].put('Foo', {})
        self.client.queues['foo'].put('Foo', {})
        self.assertEqual(len(self.client.queues['foo'].pop(10)), 2)

    def test_peek(self):
        '''Exposes queue peeking'''
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.assertEqual(self.client.queues['foo'].peek().jid, 'jid')

    def test_multipeek(self):
        '''Exposes multi-peek'''
        self.client.queues['foo'].put('Foo', {})
        self.client.queues['foo'].put('Foo', {})
        self.assertEqual(len(self.client.queues['foo'].peek(10)), 2)

    def test_stats(self):
        '''Exposes stats'''
        self.client.queues['foo'].stats()

    def test_len(self):
        '''Exposes the length of a queue'''
        self.client.queues['foo'].put('Foo', {})
        self.assertEqual(len(self.client.queues['foo']), 1)



'''A Gevent-based worker'''

import os
import gevent
import gevent.pool

from . import Worker
from qless import logger


class GeventWorker(Worker):
    '''A Gevent-based worker'''
    def __init__(self, *args, **kwargs):
        Worker.__init__(self, *args, **kwargs)
        # Should we shut down after this?
        self.shutdown = False
        # A mapping of jids to the greenlets handling them
        self.greenlets = {}
        count = kwargs.pop('greenlets', 10)
        self.pool = gevent.pool.Pool(count)
        # A list of the sandboxes that we'll use
        self.sandbox = kwargs.pop(
            'sandbox', os.path.join(os.getcwd(), 'qless-py-workers'))
        self.sandboxes = [
            os.path.join(self.sandbox, 'greenlet-%i' % i) for i in range(count)]

    def process(self, job):
        '''Process a job'''
        sandbox = self.sandboxes.pop(0)
        try:
            with Worker.sandbox(sandbox):
                job.sandbox = sandbox
                job.process()
        finally:
            # Delete its entry from our greenlets mapping
            self.greenlets.pop(job.jid, None)
            self.sandboxes.append(sandbox)

    def kill(self, jid):
        '''Stop the greenlet processing the provided jid'''
        greenlet = self.greenlets.get(jid)
        if greenlet != None:
            logger.warn('Lost ownership of %s' % jid)
            greenlet.kill()

    @classmethod
    def patch(cls):  # pragma: no cover
        '''Monkey-patch anything that needs to be patched'''
        from gevent import monkey
        monkey.patch_all()

    def run(self):
        '''Work on jobs'''
        # Register signal handlers
        self.signals()

        # And monkey-patch before doing any imports
        self.patch()

        # Start listening
        with self.listener():
            try:
                generator = self.jobs()
                while not self.shutdown:
                    self.pool.wait_available()
                    job = generator.next()
                    if job:
                        # For whatever reason, doing imports within a greenlet
                        # (there's one implicitly invoked in job.process), was
                        # throwing exceptions. The hacky way to get around this
                        # is to force the import to happen before the greenlet
                        # is spawned.
                        job.klass
                        greenlet = gevent.Greenlet(self.process, job)
                        self.greenlets[job.jid] = greenlet
                        self.pool.start(greenlet)
                    else:
                        logger.debug('Sleeping for %fs' % self.interval)
                        gevent.sleep(self.interval)
            except StopIteration:
                logger.info('Exhausted jobs')
            finally:
                logger.info('Waiting for greenlets to finish')
                self.pool.join()



#! /usr/bin/env python

'''Our base worker'''

import os
import code
import signal
import shutil
import itertools
import traceback
import threading
from contextlib import contextmanager

# Internal imports
from qless.listener import Listener
from qless import logger, exceptions

# Try to use the fast json parser
try:
    import simplejson as json
except ImportError:  # pragma: no cover
    import json

# Setting the process title
try:
    from setproctitle import setproctitle, getproctitle
except ImportError:  # pragma: no cover
    def setproctitle(title):
        pass

    def getproctitle():
        return ''


class Worker(object):
    '''Worker. For doing work'''
    @classmethod
    def title(cls, message=None):
        '''Set the title of the process'''
        if message == None:
            return getproctitle()
        else:
            setproctitle('qless-py-worker %s' % message)
            logger.info(message)

    @classmethod
    def divide(cls, jobs, count):
        '''Divide up the provided jobs into count evenly-sized groups'''
        jobs = list(zip(*itertools.izip_longest(*[iter(jobs)] * count)))
        # If we had no jobs to resume, then we get an empty list
        jobs = jobs or [()] * count
        for index in xrange(count):
            # Filter out the items in jobs that are Nones
            jobs[index] = [j for j in jobs[index] if j != None]
        return jobs

    @classmethod
    def clean(cls, path):
        '''Clean up all the files in a provided path'''
        for pth in os.listdir(path):
            pth = os.path.abspath(os.path.join(path, pth))
            if os.path.isdir(pth):
                logger.debug('Removing directory %s' % pth)
                shutil.rmtree(pth)
            else:
                logger.debug('Removing file %s' % pth)
                os.remove(pth)

    @classmethod
    @contextmanager
    def sandbox(cls, path):
        '''Ensures path exists before yielding, cleans up after'''
        # Ensure the path exists and is clean
        if not os.path.exists(path):
            logger.debug('Making %s' % path)
            os.makedirs(path)
        cls.clean(path)
        # Then yield, but make sure to clean up the directory afterwards
        try:
            yield
        finally:
            cls.clean(path)

    def __init__(self, queues, client, **kwargs):
        self.client = client
        # This should accept either queue objects, or string queue names
        self.queues = []
        for queue in queues:
            if isinstance(queue, basestring):
                self.queues.append(self.client.queues[queue])
            else:
                self.queues.append(queue)

        # Save our kwargs, since a common pattern to instantiate subworkers
        self.kwargs = kwargs
        # Check for any jobs that we should resume. If 'resume' is the actual
        # value 'True', we should find all the resumable jobs we can. Otherwise,
        # we should interpret it as a list of jobs already
        self.resume = kwargs.get('resume') or []
        if self.resume == True:
            self.resume = self.resumable()
        # How frequently we should poll for work
        self.interval = kwargs.get('interval', 60)
        # To mark whether or not we should shutdown after work is done
        self.shutdown = False

    def resumable(self):
        '''Find all the jobs that we'd previously been working on'''
        # First, find the jids of all the jobs registered to this client.
        # Then, get the corresponding job objects
        jids = self.client.workers[self.client.worker_name]['jobs']
        jobs = self.client.jobs.get(*jids)

        # We'll filter out all the jobs that aren't in any of the queues
        # we're working on.
        queue_names = set([queue.name for queue in self.queues])
        return [job for job in jobs if job.queue_name in queue_names]

    def jobs(self):
        '''Generator for all the jobs'''
        # If we should resume work, then we should hand those out first,
        # assuming we can still heartbeat them
        for job in self.resume:
            try:
                if job.heartbeat():
                    yield job
            except exceptions.LostLockException:
                logger.exception('Cannot resume %s' % job.jid)
        while True:
            seen = False
            for queue in self.queues:
                job = queue.pop()
                if job:
                    seen = True
                    yield job
            if not seen:
                yield None

    @contextmanager
    def listener(self):
        '''Listen for pubsub messages relevant to this worker in a thread'''
        channels = ['ql:w:' + self.client.worker_name]
        listener = Listener(self.client.redis, channels)
        print 'Self: %s' % self
        thread = threading.Thread(target=self.listen, args=(listener,))
        thread.start()
        try:
            yield
        finally:
            listener.unlisten()
            thread.join()

    def listen(self, listener):
        '''Listen for events that affect our ownership of a job'''
        for message in listener.listen():
            try:
                data = json.loads(message['data'])
                if data['event'] in ('canceled', 'lock_lost', 'put'):
                    self.kill(data['jid'])
            except:
                logger.exception('Pubsub error')

    def kill(self, jid):
        '''Stop processing the provided jid'''
        raise NotImplementedError('Derived classes must override "kill"')

    def signals(self, signals=('QUIT', 'USR1', 'USR2')):
        '''Register our signal handler'''
        for sig in signals:
            signal.signal(getattr(signal, 'SIG' + sig), self.handler)

    def stop(self):
        '''Mark this for shutdown'''
        self.shutdown = True

    # Unfortunately, for most of this, it's not really practical to unit test
    def handler(self, signum, frame):  # pragma: no cover
        '''Signal handler for this process'''
        if signum == signal.SIGQUIT:
            # QUIT - Finish processing, but don't do any more work after that
            self.stop()
        elif signum == signal.SIGUSR1:
            # USR1 - Print the backtrace
            message = ''.join(traceback.format_stack(frame))
            message = 'Signaled traceback for %s:\n%s' % (os.getpid(), message)
            print message
            logger.warn(message)
        elif signum == signal.SIGUSR2:
            # USR2 - Enter a debugger
            # Much thanks to http://stackoverflow.com/questions/132058
            data = {'_frame': frame}    # Allow access to frame object.
            data.update(frame.f_globals)  # Unless shadowed by global
            data.update(frame.f_locals)
            # Build up a message with a traceback
            message = ''.join(traceback.format_stack(frame))
            message = 'Traceback:\n%s' % message
            code.InteractiveConsole(data).interact(message)



'''Tests about events'''

from common import TestQless


class TestEvents(TestQless):
    '''Tests about events'''
    def setUp(self):
        TestQless.setUp(self)
        self.client.queues['foo'].put('Foo', {}, jid='jid')
        self.client.jobs['jid'].track()

    def test_basic(self):
        '''Ensure we can get a basic event'''
        def func(_):
            '''No docstring'''
            func.count += 1

        func.count = 0
        self.client.events.on('popped', func)
        with self.client.events.thread():
            self.client.queues['foo'].pop()
        self.assertEqual(func.count, 1)

    def test_off(self):
        '''Ensure we can turn off callbacks'''
        def popped(_):
            '''No docstring'''
            popped.count += 1

        def completed(_):
            '''No docstring'''
            completed.count += 1

        popped.count = 0
        completed.count = 0
        self.client.events.on('popped', popped)
        self.client.events.on('completed', completed)
        self.client.events.off('popped')
        with self.client.events.thread():
            self.client.queues['foo'].pop().complete()
        self.assertEqual(popped.count, 0)
        self.assertEqual(completed.count, 1)

    def test_not_implemented(self):
        '''Ensure missing events throw errors'''
        self.assertRaises(
            NotImplementedError, self.client.events.on, 'foo', int)
